@@ -3,22 +3,17 @@ from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union
 
 import evaluate
 import numpy as np
+import torch
 from sentence_transformers import InputExample, losses
 from sentence_transformers.datasets import SentenceLabelDataset
 from sentence_transformers.losses.BatchHardTripletLoss import BatchHardTripletLossDistanceFunction
 from torch.utils.data import DataLoader
-from transformers.trainer_utils import (
-    BestRun,
-    HPSearchBackend,
-    default_compute_objective,
-    number_of_arguments,
-    set_seed,
-)
+from transformers.trainer_utils import HPSearchBackend, default_compute_objective, number_of_arguments, set_seed
 
 from . import logging
 from .integrations import default_hp_search_backend, is_optuna_available, run_hp_search_optuna
 from .modeling import SupConLoss, sentence_pairs_generation, sentence_pairs_generation_multilabel
-from .utils import default_hp_space_optuna
+from .utils import BestRun, default_hp_space_optuna
 
 
 if TYPE_CHECKING:
@@ -43,9 +38,10 @@ class SetFitTrainer:
             The evaluation dataset.
         model_init (`Callable[[], SetFitModel]`, *optional*):
             A function that instantiates the model to be used. If provided, each call to [`~SetFitTrainer.train`] will start
-            from a new instance of the model as given by this function.
-        metric (`str`, *optional*, defaults to `"accuracy"`):
-            The metric to use for evaluation.
+            from a new instance of the model as given by this function when a `trial` is passed.
+        metric (`str` or `Callable`, *optional*, defaults to `"accuracy"`):
+            The metric to use for evaluation. If a string is provided, we treat it as the metric name and load it with default settings.
+            If a callable is provided, it must take two arguments (`y_pred`, `y_test`).
         loss_class (`nn.Module`, *optional*, defaults to `CosineSimilarityLoss`):
             The loss function to use for contrastive training.
         num_iterations (`int`, *optional*, defaults to `20`):
@@ -61,6 +57,18 @@ class SetFitTrainer:
             [`~SetTrainer.model_init`] function to instantiate the model if it has some randomly initialized parameters.
         column_mapping (`Dict[str, str]`, *optional*):
             A mapping from the column names in the dataset to the column names expected by the model. The expected format is a dictionary with the following format: {"text_column_name": "text", "label_column_name: "label"}.
+        use_amp (`bool`, *optional*, defaults to `False`):
+            Use Automatic Mixed Precision (AMP). Only for Pytorch >= 1.6.0
+        warmup_proportion (`float`, *optional*, defaults to `0.1`):
+            Proportion of the warmup in the total training steps.
+            Must be greater than or equal to 0.0 and less than or equal to 1.0.
+        distance_metric (`Callable`, defaults to `BatchHardTripletLossDistanceFunction.cosine_distance`):
+            Function that returns a distance between two embeddings.
+            It is set for the triplet loss and
+            is ignored for `CosineSimilarityLoss` and `SupConLoss`.
+        margin (`float`, defaults to `0.25`): Margin for the triplet loss.
+            Negative samples should be at least margin further apart from the anchor than the positive.
+            This is ignored for `CosineSimilarityLoss`, `BatchHardSoftMarginTripletLoss` and `SupConLoss`.
     """
 
     def __init__(
@@ -69,7 +77,7 @@ class SetFitTrainer:
         train_dataset: "Dataset" = None,
         eval_dataset: "Dataset" = None,
         model_init: Callable[[], "SetFitModel"] = None,
-        metric: str = "accuracy",
+        metric: Union[str, Callable[["Dataset", "Dataset"], Dict[str, float]]] = "accuracy",
         loss_class=losses.CosineSimilarityLoss,
         num_iterations: int = 20,
         num_epochs: int = 1,
@@ -77,7 +85,15 @@ class SetFitTrainer:
         batch_size: int = 16,
         seed: int = 42,
         column_mapping: Dict[str, str] = None,
+        use_amp: bool = False,
+        warmup_proportion: float = 0.1,
+        distance_metric: Callable = BatchHardTripletLossDistanceFunction.cosine_distance,
+        margin: float = 0.25,
     ):
+        if (warmup_proportion < 0.0) or (warmup_proportion > 1.0):
+            raise ValueError(
+                f"warmup_proportion must be greater than or equal to 0.0 and less than or equal to 1.0! But it was: {warmup_proportion}"
+            )
 
         self.train_dataset = train_dataset
         self.eval_dataset = eval_dataset
@@ -89,6 +105,10 @@ class SetFitTrainer:
         self.batch_size = batch_size
         self.seed = seed
         self.column_mapping = column_mapping
+        self.use_amp = use_amp
+        self.warmup_proportion = warmup_proportion
+        self.distance_metric = distance_metric
+        self.margin = margin
 
         if model is None:
             if model_init is not None:
@@ -104,12 +124,13 @@ class SetFitTrainer:
 
         self.model = model
         self.hp_search_backend = None
+        self._freeze = True  # If True, will train the body only; otherwise, train the body and head
 
     def _validate_column_mapping(self, dataset: "Dataset") -> None:
         """
         Validates the provided column mapping against the dataset.
         """
-        required_columns = set(["text", "label"])
+        required_columns = {"text", "label"}
         column_names = set(dataset.column_names)
         if self.column_mapping is None and not required_columns.issubset(column_names):
             raise ValueError(
@@ -146,40 +167,53 @@ class SetFitTrainer:
         )
         return dataset
 
+    def apply_hyperparameters(self, params: Dict[str, Any], final_model: bool = False):
+        """Applies a dictionary of hyperparameters to both the trainer and the model
+
+        Args:
+            params (`Dict[str, Any]`): The parameters, usually from `BestRun.hyperparameters`
+            final_model (`bool`, *optional*, defaults to `False`): If `True`, replace the `model_init()` function with a fixed model based on the parameters.
+        """
+        for key, value in params.items():
+            if hasattr(self, key):
+                old_attr = getattr(self, key, None)
+                # Casting value to the proper type
+                if old_attr is not None:
+                    value = type(old_attr)(value)
+                setattr(self, key, value)
+            elif number_of_arguments(self.model_init) == 0:  # we do not warn if model_init could be using it
+                logger.warning(
+                    f"Trying to set {key} in the hyperparameter search but there is no corresponding field in "
+                    "`SetFitTrainer`, and `model_init` does not take any arguments."
+                )
+
+        self.model = self.model_init(params)
+        if final_model:
+            self.model_init = None
+
     def _hp_search_setup(self, trial: Union["optuna.Trial", Dict[str, Any]]):
         """HP search setup code"""
 
         # Heavily inspired by transformers.Trainer._hp_search_setup
-        # https://github.com/huggingface/transformers/blob/cbb8a37929c3860210f95c9ec99b8b84b8cf57a1/src/transformers/trainer.py#L1163        self._trial = trial
-        self._trial = trial
-
         if self.hp_search_backend is None or trial is None:
             return
-        if self.hp_search_backend == HPSearchBackend.OPTUNA:
+
+        if isinstance(trial, Dict):  # For passing a Dict to train() -- mostly unused for now
+            params = trial
+        elif self.hp_search_backend == HPSearchBackend.OPTUNA:
             params = self.hp_space(trial)
+        else:
+            raise ValueError("Invalid trial parameter")
 
-        for key, value in params.items():
-            if not hasattr(self, key):
-                logger.warning(
-                    f"Trying to set {key} in the hyperparameter search but there is no corresponding field in"
-                    " `SetFitTrainer`."
-                )
-                continue
-            old_attr = getattr(self, key, None)
-            # Casting value to the proper type
-            if old_attr is not None:
-                value = type(old_attr)(value)
-            setattr(self, key, value)
+        logger.info(f"Trial: {params}")
+        self.apply_hyperparameters(params, final_model=False)
 
-        if self.hp_search_backend == HPSearchBackend.OPTUNA:
-            logger.info("Trial:", trial.params)
-
-    def call_model_init(self, trial=None):
+    def call_model_init(self, params: Dict[str, Any] = None):
         model_init_argcount = number_of_arguments(self.model_init)
         if model_init_argcount == 0:
             model = self.model_init()
         elif model_init_argcount == 1:
-            model = self.model_init(trial)
+            model = self.model_init(params)
         else:
             raise RuntimeError("model_init should have 0 or 1 argument.")
 
@@ -188,20 +222,69 @@ class SetFitTrainer:
 
         return model
 
-    def train(self, trial: Union["optuna.Trial", Dict[str, Any]] = None):
+    def freeze(self):
+        """
+        Freeze SetFitModel's differentiable head.
+        Note: call this function only when using the differentiable head.
+        """
+        if not isinstance(self.model.model_head, torch.nn.Module):
+            raise ValueError("Please use the differentiable head in `SetFitModel` when calling this function.")
+
+        self._freeze = True  # Currently use self._freeze as a switch
+        self.model.freeze("head")
+
+    def unfreeze(self, keep_body_frozen: bool = False):
+        """
+        Unfreeze SetFitModel's differentiable head.
+        Note: call this function only when using the differentiable head.
+
+        Args:
+            keep_body_frozen (`bool`, *optional*, defaults to `False`):
+                Whether to freeze the body when unfreeze the head.
+        """
+        if not isinstance(self.model.model_head, torch.nn.Module):
+            raise ValueError("Please use the differentiable head in `SetFitModel` when calling this function.")
+
+        self._freeze = False  # Currently use self._freeze as a switch
+        self.model.unfreeze("head")
+        if keep_body_frozen:
+            self.model.freeze("body")
+        else:  # ensure to unfreeze the body
+            self.model.unfreeze("body")
+
+    def train(
+        self,
+        num_epochs: Optional[int] = None,
+        batch_size: Optional[int] = None,
+        learning_rate: Optional[float] = None,
+        body_learning_rate: Optional[float] = None,
+        l2_weight: Optional[float] = None,
+        trial: Union["optuna.Trial", Dict[str, Any]] = None,
+    ):
         """
         Main training entry point.
 
         Args:
+            num_epochs (int, *optional*):
+                Temporary change the number of epochs to train the Sentence Transformer body/head for.
+                If ignore, will use the value given in initialization.
+            batch_size (int, *optional*):
+                Temporary change the batch size to use for contrastive training or logistic regression.
+                If ignore, will use the value given in initialization.
+            learning_rate (float, *optional*):
+                Temporary change the learning rate to use for contrastive training or SetFitModel's head in logistic regression.
+                If ignore, will use the value given in initialization.
+            body_learning_rate (float, *optional*):
+                Temporary change the learning rate to use for SetFitModel's body in logistic regression only.
+                If ignore, will be the same as `learning_rate`.
+            l2_weight (float, *optional*):
+                Temporary change the weight of L2 regularization for SetFitModel's differentiable head in logistic regression.
             trial (`optuna.Trial` or `Dict[str, Any]`, *optional*):
                 The trial run or the hyperparameter dictionary for hyperparameter search.
         """
-        self._hp_search_setup(trial)
-        # Model re-init
-        if self.model_init is not None:
-            # Seed must be set before instantiating the model when using model_init.
-            set_seed(self.seed)
-            self.model = self.call_model_init(trial)
+        if trial:  # Trial and model initialization
+            set_seed(self.seed)  # Seed must be set before instantiating the model when using model_init.
+            self._hp_search_setup(trial)  # sets trainer parameters and initializes model
 
         if self.train_dataset is None:
             raise ValueError("SetFitTrainer: training requires a train_dataset.")
@@ -211,99 +294,134 @@ class SetFitTrainer:
         if self.column_mapping is not None:
             logger.info("Applying column mapping to training dataset")
             train_dataset = self._apply_column_mapping(self.train_dataset, self.column_mapping)
+
         x_train = train_dataset["text"]
         y_train = train_dataset["label"]
-
         if self.loss_class is None:
             return
 
-        # sentence-transformers adaptation
-        batch_size = self.batch_size
-        if self.loss_class in [
-            losses.BatchAllTripletLoss,
-            losses.BatchHardTripletLoss,
-            losses.BatchSemiHardTripletLoss,
-            losses.BatchHardSoftMarginTripletLoss,
-            SupConLoss,
-        ]:
-            train_examples = [InputExample(texts=[text], label=label) for text, label in zip(x_train, y_train)]
-            train_data_sampler = SentenceLabelDataset(train_examples)
+        num_epochs = num_epochs or self.num_epochs
+        batch_size = batch_size or self.batch_size
+        learning_rate = learning_rate or self.learning_rate
+        batch_size = batch_size or self.batch_size
+        is_differentiable_head = isinstance(self.model.model_head, torch.nn.Module)  # If False, assume using sklearn
 
-            batch_size = min(self.batch_size, len(train_data_sampler))
-            train_dataloader = DataLoader(train_data_sampler, batch_size=batch_size, drop_last=True)
+        if not is_differentiable_head or self._freeze:
+            # sentence-transformers adaptation
+            if self.loss_class in [
+                losses.BatchAllTripletLoss,
+                losses.BatchHardTripletLoss,
+                losses.BatchSemiHardTripletLoss,
+                losses.BatchHardSoftMarginTripletLoss,
+                SupConLoss,
+            ]:
+                train_examples = [InputExample(texts=[text], label=label) for text, label in zip(x_train, y_train)]
+                train_data_sampler = SentenceLabelDataset(train_examples)
 
-            if self.loss_class is losses.BatchHardSoftMarginTripletLoss:
-                train_loss = self.loss_class(
-                    model=self.model,
-                    distance_metric=BatchHardTripletLossDistanceFunction.cosine_distance,
-                )
-            elif self.loss_class is SupConLoss:
-                train_loss = self.loss_class(model=self.model)
-            else:
-                train_loss = self.loss_class(
-                    model=self.model,
-                    distance_metric=BatchHardTripletLossDistanceFunction.cosine_distance,
-                    margin=0.25,
-                )
+                batch_size = min(batch_size, len(train_data_sampler))
+                train_dataloader = DataLoader(train_data_sampler, batch_size=batch_size, drop_last=True)
 
-            train_steps = len(train_dataloader) * self.num_epochs
-        else:
-            train_examples = []
-
-            for _ in range(self.num_iterations):
-                if self.model.multi_target_strategy is not None:
-                    train_examples = sentence_pairs_generation_multilabel(
-                        np.array(x_train), np.array(y_train), train_examples
+                if self.loss_class is losses.BatchHardSoftMarginTripletLoss:
+                    train_loss = self.loss_class(
+                        model=self.model.model_body,
+                        distance_metric=self.distance_metric,
                     )
+                elif self.loss_class is SupConLoss:
+                    train_loss = self.loss_class(model=self.model.model_body)
                 else:
-                    train_examples = sentence_pairs_generation(np.array(x_train), np.array(y_train), train_examples)
+                    train_loss = self.loss_class(
+                        model=self.model.model_body,
+                        distance_metric=self.distance_metric,
+                        margin=self.margin,
+                    )
 
-            train_dataloader = DataLoader(train_examples, shuffle=True, batch_size=self.batch_size)
-            train_loss = self.loss_class(self.model.model_body)
-            train_steps = len(train_dataloader)
+                train_steps = len(train_dataloader) * self.num_epochs
+            else:
+                train_examples = []
 
-        logger.info("***** Running training *****")
-        logger.info(f"  Num examples = {len(train_examples)}")
-        logger.info(f"  Num epochs = {self.num_epochs}")
-        logger.info(f"  Total optimization steps = {train_steps}")
-        logger.info(f"  Total train batch size = {batch_size}")
+                for _ in range(self.num_iterations):
+                    if self.model.multi_target_strategy is not None:
+                        train_examples = sentence_pairs_generation_multilabel(
+                            np.array(x_train), np.array(y_train), train_examples
+                        )
+                    else:
+                        train_examples = sentence_pairs_generation(
+                            np.array(x_train), np.array(y_train), train_examples
+                        )
 
-        warmup_steps = math.ceil(train_steps * 0.1)
-        self.model.model_body.fit(
-            train_objectives=[(train_dataloader, train_loss)],
-            epochs=self.num_epochs,
-            steps_per_epoch=train_steps,
-            optimizer_params={"lr": self.learning_rate},
-            warmup_steps=warmup_steps,
-            show_progress_bar=True,
-        )
+                train_dataloader = DataLoader(train_examples, shuffle=True, batch_size=batch_size)
+                train_loss = self.loss_class(self.model.model_body)
+                train_steps = len(train_dataloader) * num_epochs
 
-        # Train the final classifier
-        self.model.fit(x_train, y_train)
+            logger.info("***** Running training *****")
+            logger.info(f"  Num examples = {len(train_examples)}")
+            logger.info(f"  Num epochs = {num_epochs}")
+            logger.info(f"  Total optimization steps = {train_steps}")
+            logger.info(f"  Total train batch size = {batch_size}")
+
+            warmup_steps = math.ceil(train_steps * self.warmup_proportion)
+            self.model.model_body.fit(
+                train_objectives=[(train_dataloader, train_loss)],
+                epochs=num_epochs,
+                steps_per_epoch=train_steps,
+                optimizer_params={"lr": learning_rate},
+                warmup_steps=warmup_steps,
+                show_progress_bar=True,
+                use_amp=self.use_amp,
+            )
+
+        if not is_differentiable_head or not self._freeze:
+            # Train the final classifier
+            self.model.fit(
+                x_train,
+                y_train,
+                num_epochs=num_epochs,
+                batch_size=batch_size,
+                learning_rate=learning_rate,
+                body_learning_rate=body_learning_rate,
+                l2_weight=l2_weight,
+                show_progress_bar=True,
+            )
 
     def evaluate(self):
-        """Computes the metrics for a given classifier."""
+        """
+        Computes the metrics for a given classifier.
+
+        Returns:
+            `Dict[str, float]`: The evaluation metrics.
+        """
+
         self._validate_column_mapping(self.eval_dataset)
         eval_dataset = self.eval_dataset
+
         if self.column_mapping is not None:
             logger.info("Applying column mapping to evaluation dataset")
             eval_dataset = self._apply_column_mapping(self.eval_dataset, self.column_mapping)
-        metric_config = "multilabel" if self.model.multi_target_strategy is not None else None
-        metric_fn = evaluate.load(self.metric, config_name=metric_config)
+
         x_test = eval_dataset["text"]
         y_test = eval_dataset["label"]
 
         logger.info("***** Running evaluation *****")
         y_pred = self.model.predict(x_test)
 
-        return metric_fn.compute(predictions=y_pred, references=y_test)
+        if isinstance(self.metric, str):
+            metric_config = "multilabel" if self.model.multi_target_strategy is not None else None
+            metric_fn = evaluate.load(self.metric, config_name=metric_config)
+
+            return metric_fn.compute(predictions=y_pred, references=y_test)
+
+        elif callable(self.metric):
+            return self.metric(y_pred, y_test)
+
+        else:
+            raise ValueError("metric must be a string or a callable")
 
     def hyperparameter_search(
         self,
         hp_space: Optional[Callable[["optuna.Trial"], Dict[str, float]]] = None,
         compute_objective: Optional[Callable[[Dict[str, float]], float]] = None,
         n_trials: int = 10,
-        direction: str = "minimize",
+        direction: str = "maximize",
         backend: Optional[Union["str", HPSearchBackend]] = None,
         hp_name: Optional[Callable[["optuna.Trial"], str]] = None,
         **kwargs,
@@ -326,10 +444,10 @@ class SetFitTrainer:
                 [`~trainer_utils.default_hp_space_optuna`].
             compute_objective (`Callable[[Dict[str, float]], float]`, *optional*):
                 A function computing the objective to minimize or maximize from the metrics returned by the `evaluate`
-                method. Will default to [`~trainer_utils.default_compute_objective`].
+                method. Will default to [`~trainer_utils.default_compute_objective`] which uses the sum of metrics.
             n_trials (`int`, *optional*, defaults to 100):
                 The number of trial runs to test.
-            direction (`str`, *optional*, defaults to `"minimize"`):
+            direction (`str`, *optional*, defaults to `"maximize"`):
                 Whether to optimize greater or lower objects. Can be `"minimize"` or `"maximize"`, you should pick
                 `"minimize"` when optimizing the validation loss, `"maximize"` when optimizing one or several metrics.
             backend (`str` or [`~training_utils.HPSearchBackend`], *optional*):
